/***
 * JRSFuzz
 * by Javantea
 * Oct 11, 2015
 * 
 * JRSFuzz is a dumb fuzzer with a goal of fuzzing each byte as if it were a 
 * type in a TLV. JRSFuzz fuzzes each byte separately leaving the other 
 * bytes alone. One obvious benefit is the number of rounds is easy to 
 * compute:
 * 256 * len(data)
 * Therefore if you give it a number higher, it will return 0.
 * You must free the return value when you are finished.
 * The values tested for b'a:' are:
 * [b'\x00:', b'\x01:', b'\x02:', b'\x03:', ..., 
 *  b'\xfd:', b'\xfe:', b'\xff:', 
 *  b'a\x00', b'a\x01', b'a\x02', b'a\x03', ..., 
 *  b'a\xfd', b'a\xfe', b'a\xff']
 * As you can see with the example, this is an excellent short grammar 
 * fuzzer, working better than most smart fuzzers. However, it lacks the 
 * intelligence and finess of most smart fuzzers, not changing the length 
 * of any value.
 * Another benefit is the scaling. With data length 5, it has 1280 rounds. 
 * With data length 5000, it has 1280000 rounds which is totally 
 * reasonable for a dumb fuzzer.
 * 
 * An inline version of this fuzzer could exist and be much more efficient, 
 * but I haven't written it to avoid confusion.
 * 
 ***/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "jrsfuzz.h"

char *JRSFuzz(char *data, size_t data_len, int i)
{
	if(i < 0)
	{
		return 0;
	}
	char *ret = malloc(data_len);
	if(ret == 0)
	{
		return 0;
	}
	memcpy(ret, data, data_len);
	// This is actually >> 8 and & 0xff, very cheap.
	// 4007f0: SAR EAX, 0x8 ; EAX >>= 0x8; 
	int pos = i / 256;
	// 4007fa: MOV [RCX+RAX], BL ; [RCX+RAX] = BL; 
	int v = i % 256;
	if(pos >= data_len)
	{
		free(ret);
		return 0;
	}
	ret[pos] = v;
	return ret;
}

#ifdef TEST

#include <stdio.h>
#include <unistd.h>

/*
 * Run the fuzzer on a string you provide.
 * If you want to use nulls on the input or want a large input, use the file 
 * version.
 * 
 */
int main(int argc, char **argv)
{
	char *data = "a:";
	int data_len = 2;
	if(argc > 1)
	{
		data = argv[1];
		data_len = strlen(data);
	}
	// TODO: Test on 32, 64 diff archs and so on.
	int i;
	for(i = 0; i < 256 * data_len; i++)
	{
		char *r = JRSFuzz(data, data_len, i);
		if(r == 0)
		{
			perror("JRSFuzz");
			return 1;
		}
		fwrite(r, data_len, 1, stdout);
		free(r);
	}
	return 0;
}

#endif // TEST
