#!/usr/bin/env python3
"""
PNG Fuzzer using JRSFuzz
by Javantea
Nov 14, 2015
PDF Fuzzer using JRSFuzz
by Javantea
Oct 29, 2015
DNS Fuzzer using JRSFuzz
by Javantea
Oct 11, 2015
Based on SMTP Fuzzer using JRSFuzz
by Javantea
Oct 11, 2015
Based on SMTP Grammar Fuzzer
by Javantea
Sept 14, 2015

A very reasonable dumb fuzzer for PNG. It's a mutation fuzzer.
Provide the following values to fuzz a binary file you have captured:
python3 png_jrsfuzz1.py filename

TIME='%M kB max memory\nReal:%e seconds\n' time pngtopnm bigsharp.png >/dev/null
588900 kB max memory
Real:9.30 seconds

TODO:
Fix CRC32 using parser.
"""
import jrsfuzz
import sys
import debug1
import time
import tempfile
import zlib
import struct

def parsePNG(f, fpos=0):
	"""
	Given N bytes of data that passes findPNG, find the end and parts of a 
	PNG.
	PNG format is:
	magic \x89PNG
	0d 0a 1a 0a
	chunk length 00 00 00 0d
	IHDR
	Width 	4 bytes
	Height 	4 bytes
	Bit depth 	1 byte
	Colour type 	1 byte
	Compression method 	1 byte
	Filter method 	1 byte
	Interlace method 	1 byte
	"""
	flen = len(f) - fpos
	out = {'sig':f[fpos:fpos+4]}
	if flen < 8:
		return out
	#end if
	out['sig2'] = f[fpos+4:fpos+8]
	if flen < 12:
		return out
	#end if
	IHDR_len, = struct.unpack('>I', f[fpos+8:fpos+12])
	out['IHDR_len'] = IHDR_len
	if flen < 16:
		return out
	#end if
	out['IHDR'] = f[fpos+12:fpos+16]
	if flen < 29:
		return out
	#end if
	width, = struct.unpack('>I', f[fpos+16:fpos+20])
	out['width']  = width
	height, = struct.unpack('>I', f[fpos+20:fpos+24])
	out['height'] = height
	bit_depth = (f[fpos+24])
	out['bit_depth'] = bit_depth
	color_type = (f[fpos+25])
	out['color_type'] = color_type
	compression_method = (f[fpos+26])
	out['compression_method'] = compression_method
	filter_method = (f[fpos+27])
	out['filter_method'] = filter_method
	interlace_method = (f[fpos+28])
	out['interlace_method'] = interlace_method
	if flen < 33:
		return out
	#end if	
	IHDR_crc32, = struct.unpack('>i', f[fpos+29:fpos+33])
	out['IHDR_crc32'] = IHDR_crc32
	IHDR_cmp_crc32 = zlib.crc32(f[fpos+12:fpos+29])
	out['IHDR_cmp_crc32'] = IHDR_cmp_crc32
	if IHDR_crc32 != IHDR_cmp_crc32:
		out['Error'] = ['IHDR crc32 does not match']
	#end if
	out['chunks'] = []
	cpos = fpos+33
	ctype = 'IHDR'
	while ctype != 'IEND':
		if flen < cpos+4: break
		clen, = struct.unpack('>I', f[cpos:cpos+4])
		if flen < cpos+8+clen+4: break
		ctype = f[cpos+4:cpos+8]
		cdata = f[cpos+8:cpos+8+clen]
		# NOTE: originally was signed. Not sure why.
		c_crc32, = struct.unpack('>I', f[cpos+8+clen:cpos+8+clen+4])
		actual_crc32 = zlib.crc32(f[cpos+4:cpos+8+clen])
		out['chunks'].append((clen, ctype, cdata, c_crc32, actual_crc32))
		cpos += 8 + clen + 4
	#loop
	out['end'] = cpos
	return out
#end def parsePNG(f, fpos)

def main():
	# This works on any PNG file
	data = b'%'
	valgrind = False
	if '-v' in sys.argv:
		valgrind = True
		sys.argv.remove('-v')
	#end if
	if len(sys.argv) > 1:
		filename = sys.argv[1]
		data = open(filename, 'rb').read()
	#end if
	lines_output = len(data) * 256
	print("%i outputs, %3.3f MB" % (lines_output, lines_output*len(data)/(1<<20)), file=sys.stderr)
	start, end = 0, lines_output
	if len(sys.argv) > 2:
		start = int(sys.argv[2])
	#end if
	if len(sys.argv) > 3:
		end = int(sys.argv[3])
	#end if
	if start > lines_output: start = lines_output-1
	if end > lines_output: end = lines_output
	outfile = tempfile.NamedTemporaryFile(suffix='.png', prefix='fuzzfile', dir='.', delete=False)
	filename_out = outfile.name
	
	parsed = parsePNG(data)

	#prevResp = None
	for i in range(start, end):
		x = jrsfuzz.JRSFuzz(data, i)
		outfile.seek(0)
		outfile.write(x)
		outfile.flush()
		print(i)
		start = time.time()
		program = '/usr/bin/pngtopnm'
		add_args = []
		if valgrind:
			program = '/usr/bin/valgrind'
			add_args = ['valgrind']
		#end if
		r = debug1.fork_exec(program, add_args + ['/usr/bin/pngtopnm', filename_out], devNull=True)
		end = time.time()
		if r == None:
			print("Failed to execute apparently.")
			continue
		#end if
		pid, sts = r
		print("Now I have control again.", (pid, sts))
		if sts == 11:
			print("Segmentation fault on iteration {}".format(i))
		#end if
		if (end-start) >= 3:
			print("Slow execution on iteration {}".format(i))
		#end if
	#next i
#end def main()

if __name__ == '__main__':
	main()
#end if
